#include <stdio.h>
#include <stdlib.h>
#include <stdbool.h>
#include <time.h>
#include <avr/io.h>
#include <util/delay.h>
#include "nokia5110.h"

#define TITLE_X 25
#define TITLE_Y 10
#define DISPLAY_LEVEL_DELAY 2000
#define LEVEL_1_DELAY 3000
#define SCORE 5
#define LEVEL_2_DELAY 2000
#define X 25
#define Y 24
#define OFFSET 10

enum piece {
	no_piece = -1,
	i = 0,
	j = 1,
	l = 2,
	o = 3,
	s = 4,
	t = 5,
	z = 6
};

enum button {
	unpressed = -1,
	left = 0,
	right = 1,
	rotate = 2,
	speed_up = 3,
	start = 4
};

enum dir {
	none = -1,
	north = 0,
	east = 1,
	south = 2,
	west = 3
};

typedef struct {
	int x, y;
} Point;

static bool map[Y][X];
static int game_started = 0, game_over = 0, level_choosed = 0, level = 0, score = 0, fall = 0, button = unpressed, piece = no_piece, dir = none;
static Point coord[4], temp_coord[4];

static void init() {
	
	DDRA &= ~((1 << PA0) | (1 << PA1) | (1 << PA2) | (1 << PA3) | (1 << PA4));
	PORTA |= ((1 << PA0) | (1 << PA1) | (1 << PA2) | (1 << PA3) | (1 << PA4));
}

static void reset() {
	
	for (int i = 0; i < Y; ++i)
		for (int j = 0; j < X; ++j)
			map[i][j] = false;
	for (int i = 0; i < 4; ++i) {
		coord[i].x = 0;
		coord[i].y = 0;
		temp_coord[i].x = 0;
		temp_coord[i].y = 0;
	}
	game_started = 0;
	game_over = 0;
	level = 0;
	level_choosed = 0;
	score = 0;
	fall = 0;
	button = unpressed;
	piece = no_piece;
	dir = none;
}

static void start_lcd() {
	
	nokia_lcd_set_cursor(TITLE_X, TITLE_Y);
	nokia_lcd_write_string("TETRIS", 1);
	nokia_lcd_set_cursor(0, 20);
	nokia_lcd_write_string("Press START to begin", 1);
	nokia_lcd_set_cursor(0, 0);
}

static void chose_level() {
	
	if (button == rotate) {
		nokia_lcd_set_cursor(TITLE_X, TITLE_Y);
		nokia_lcd_write_string("Level 1", 1);
		level = 1;
		} else if (button == speed_up) {
		nokia_lcd_set_cursor(TITLE_X, TITLE_Y);
		nokia_lcd_write_string("Level 2", 1);
		level = 2;
	}
}

static void game_over_lcd() {
	
	nokia_lcd_set_cursor(TITLE_X, TITLE_Y);
	nokia_lcd_write_string("You lost!", 1);
	nokia_lcd_set_cursor(0, 20);
	nokia_lcd_write_string((const char*) score, 1);
	_delay_ms(10000);
	nokia_lcd_set_cursor(0, 0);
}

static void display_map() {
	
	for (int i = 0; i < Y; ++i) {
		nokia_lcd_set_pixel(2 * OFFSET - 2, 2 * i, 1);
		nokia_lcd_set_pixel(2 * OFFSET - 2, 2 * i + 1, 1);
		nokia_lcd_set_pixel(2 * OFFSET + 2 * X + 2, 2 * i, 1);
		nokia_lcd_set_pixel(2 * OFFSET + 2 * X + 2, 2 * i + 1, 1);
		for (int j = 0; j < X; ++j) {
			if (map[i][j]) {
				nokia_lcd_set_pixel(2 * j + 2 * OFFSET, 2 * i, 1);
				nokia_lcd_set_pixel(2 * j + 1 + 2 * OFFSET, 2 * i, 1);
				nokia_lcd_set_pixel(2 * j + 2 * OFFSET, 2 * i + 1, 1);
				nokia_lcd_set_pixel(2 * j + 1 + 2 * OFFSET, 2 * i + 1, 1);
			}
		}
	}
}

static void get_pressed_button() {
	
	if (!(PINA & (1 << PA0)))
		button = start;
	if (!(PINA & (1 << PA1)))
		button = left;
	if (!(PINA & (1 << PA2)))
		button = right;
	if (!(PINA & (1 << PA3)))
		button = rotate;
	if (!(PINA & (1 << PA4)))
		button = speed_up;
}

static void set_piece() {
	
	for (int i = 0; i < 4; ++i)
		map[coord[i].y][coord[i].x] = true;
}

static void remove_piece() {
	
	for (int i = 0; i < 4; ++i)
		map[coord[i].y][coord[i].x] = false;
}

static void init_piece() {
	
	piece = rand() % 7;
	if (piece == i) {
		coord[0].x = 8;
		coord[0].y = 1;
		coord[1].x = 9;
		coord[1].y = 1;
		coord[2].x = 10;
		coord[2].y = 1;
		coord[3].x = 11;
		coord[3].y = 1;
	} else if (piece == j) {
		coord[0].x = 8;
		coord[0].y = 1;
		coord[1].x = 9;
		coord[1].y = 1;
		coord[2].x = 10;
		coord[2].y = 1;
		coord[3].x = 10;
		coord[3].y = 2;
	} else if (piece == l) {
		coord[0].x = 8;
		coord[0].y = 1;
		coord[1].x = 9;
		coord[1].y = 1;
		coord[2].x = 10;
		coord[2].y = 1;
		coord[3].x = 8;
		coord[3].y = 2;
	} else if (piece == o) {
		coord[0].x = 8;
		coord[0].y = 0;
		coord[1].x = 9;
		coord[1].y = 0;
		coord[2].x = 8;
		coord[2].y = 1;
		coord[3].x = 9;
		coord[3].y = 1;
	} else if (piece == s) {
		coord[0].x = 8;
		coord[0].y = 2;
		coord[1].x = 9;
		coord[1].y = 2;
		coord[2].x = 9;
		coord[2].y = 1;
		coord[3].x = 10;
		coord[3].y = 1;
	} else if (piece == t) {
		coord[0].x = 8;
		coord[0].y = 1;
		coord[1].x = 9;
		coord[1].y = 1;
		coord[2].x = 10;
		coord[2].y = 1;
		coord[3].x = 9;
		coord[3].y = 2;
	} else {
		coord[0].x = 8;
		coord[0].y = 1;
		coord[1].x = 9;
		coord[1].y = 1;
		coord[2].x = 9;
		coord[2].y = 2;
		coord[3].x = 10;
		coord[3].y = 2;
	}
	set_piece();
	dir = east;
}

static int collision() {
	
	int collision = 0;
	remove_piece();
	for (int i = 0; i < 4 && !collision; ++i)
		if (temp_coord[i].x < 0 || temp_coord[i].x >= X
			|| temp_coord[i].y < 0 || temp_coord[i].y >= Y
			|| map[temp_coord[i].y][temp_coord[i].x])
			collision = 1;
	set_piece();
	return collision;
}

static void update_coord() {
	
	remove_piece();
	for (int i = 0; i < 4; ++i) {
		coord[i].x = temp_coord[i].x;
		coord[i].y = temp_coord[i].y;
	}
	set_piece();
}

static void move_piece_to_left() {
	
	for (int i = 0; i < 4; ++i) {
		temp_coord[i].x = coord[i].x - 1;
		temp_coord[i].y = coord[i].y;
	}
	if (collision())
		return;
	update_coord();
}

static void move_piece_to_right() {
	
	for (int i = 0; i < 4; ++i) {
		temp_coord[i].x = coord[i].x + 1;
		temp_coord[i].y = coord[i].y;
	}
	if (collision())
		return;
	update_coord();
}

static void update_dir() {
	
	if (dir == north)
		dir = west;
	else
		dir--;
}

static void rotate_i() {
	
	if (dir == east) {
		temp_coord[0].x = coord[0].x + 1;
		temp_coord[0].y = coord[0].y + 2;
		temp_coord[1].x = coord[0].x + 1;
		temp_coord[1].y = coord[0].y + 1;
		temp_coord[2].x = coord[0].x + 1;
		temp_coord[2].y = coord[0].y;
		temp_coord[3].x = coord[0].x + 1;
		temp_coord[3].y = coord[0].y - 1;
	} else if (dir == north) {
		temp_coord[0].x = coord[0].x + 2;
		temp_coord[0].y = coord[0].y - 1;
		temp_coord[1].x = coord[0].x + 1;
		temp_coord[1].y = coord[0].y - 1;
		temp_coord[2].x = coord[0].x;
		temp_coord[2].y = coord[0].y - 1;
		temp_coord[3].x = coord[0].x - 1;
		temp_coord[3].y = coord[0].y - 1;
	} else if (dir == west) {
		temp_coord[0].x = coord[0].x - 1;
		temp_coord[0].y = coord[0].y - 2;
		temp_coord[1].x = coord[0].x - 1;
		temp_coord[1].y = coord[0].y - 1;
		temp_coord[2].x = coord[0].x - 1;
		temp_coord[2].y = coord[0].y;
		temp_coord[3].x = coord[0].x - 1;
		temp_coord[3].y = coord[0].y + 1;
	} else {
		temp_coord[0].x = coord[0].x - 2;
		temp_coord[0].y = coord[0].y + 1;
		temp_coord[1].x = coord[0].x - 1;
		temp_coord[1].y = coord[0].y + 1;
		temp_coord[2].x = coord[0].x;
		temp_coord[2].y = coord[0].y + 1;
		temp_coord[3].x = coord[0].x + 1;
		temp_coord[3].y = coord[0].y + 1;
	}
}

static void rotate_j() {

	if (dir == east) {
		temp_coord[0].x = coord[0].x + 1;
		temp_coord[0].y = coord[0].y + 1;
		temp_coord[1].x = coord[0].x + 1;
		temp_coord[1].y = coord[0].y;
		temp_coord[2].x = coord[0].x + 1;
		temp_coord[2].y = coord[0].y - 1;
		temp_coord[3].x = coord[0].x + 2;
		temp_coord[3].y = coord[0].y - 1;
	} else if (dir == north) {
		temp_coord[0].x = coord[0].x + 1;
		temp_coord[0].y = coord[0].y - 1;
		temp_coord[1].x = coord[0].x;
		temp_coord[1].y = coord[0].y - 1;
		temp_coord[2].x = coord[0].x - 1;
		temp_coord[2].y = coord[0].y - 1;
		temp_coord[3].x = coord[0].x - 1;
		temp_coord[3].y = coord[0].y - 2;
	} else if (dir == west) {
		temp_coord[0].x = coord[0].x - 1;
		temp_coord[0].y = coord[0].y - 1;
		temp_coord[1].x = coord[0].x - 1;
		temp_coord[1].y = coord[0].y;
		temp_coord[2].x = coord[0].x - 1;
		temp_coord[2].y = coord[0].y + 1;
		temp_coord[3].x = coord[0].x - 2;
		temp_coord[3].y = coord[0].y + 1;
	} else {
		temp_coord[0].x = coord[0].x - 1;
		temp_coord[0].y = coord[0].y + 1;
		temp_coord[1].x = coord[0].x;
		temp_coord[1].y = coord[0].y + 1;
		temp_coord[2].x = coord[0].x + 1;
		temp_coord[2].y = coord[0].y + 1;
		temp_coord[3].x = coord[0].x + 1;
		temp_coord[3].y = coord[0].y + 2;
	}
}

static void rotate_l() {
	
	if (dir == east) {
		temp_coord[0].x = coord[0].x + 1;
		temp_coord[0].y = coord[0].y + 1;
		temp_coord[1].x = coord[0].x + 1;
		temp_coord[1].y = coord[0].y;
		temp_coord[2].x = coord[0].x + 1;
		temp_coord[2].y = coord[0].y - 1;
		temp_coord[3].x = coord[0].x + 2;
		temp_coord[3].y = coord[0].y + 1;
	} else if (dir == north) {
		temp_coord[0].x = coord[0].x + 1;
		temp_coord[0].y = coord[0].y - 1;
		temp_coord[1].x = coord[0].x;
		temp_coord[1].y = coord[0].y - 1;
		temp_coord[2].x = coord[0].x - 1;
		temp_coord[2].y = coord[0].y - 1;
		temp_coord[3].x = coord[0].x + 1;
		temp_coord[3].y = coord[0].y - 2;
	} else if (dir == west) {
		temp_coord[0].x = coord[0].x - 1;
		temp_coord[0].y = coord[0].y - 1;
		temp_coord[1].x = coord[0].x - 1;
		temp_coord[1].y = coord[0].y;
		temp_coord[2].x = coord[0].x - 1;
		temp_coord[2].y = coord[0].y + 1;
		temp_coord[3].x = coord[0].x - 2;
		temp_coord[3].y = coord[0].y - 1;
	} else {
		temp_coord[0].x = coord[0].x - 1;
		temp_coord[0].y = coord[0].y + 1;
		temp_coord[1].x = coord[0].x;
		temp_coord[1].y = coord[0].y + 1;
		temp_coord[2].x = coord[0].x + 1;
		temp_coord[2].y = coord[0].y + 1;
		temp_coord[3].x = coord[0].x - 1;
		temp_coord[3].y = coord[0].y + 2;
	}
}

static void rotate_s() {
	
	if (dir == east) {
		temp_coord[0].x = coord[0].x + 2;
		temp_coord[0].y = coord[0].y;
		temp_coord[1].x = coord[0].x + 2;
		temp_coord[1].y = coord[0].y - 1;
		temp_coord[2].x = coord[0].x + 1;
		temp_coord[2].y = coord[0].y - 1;
		temp_coord[3].x = coord[0].x + 1;
		temp_coord[3].y = coord[0].y - 2;
	} else if (dir == north) {
		temp_coord[0].x = coord[0].x;
		temp_coord[0].y = coord[0].y - 2;
		temp_coord[1].x = coord[0].x - 1;
		temp_coord[1].y = coord[0].y - 2;
		temp_coord[2].x = coord[0].x - 1;
		temp_coord[2].y = coord[0].y - 1;
		temp_coord[3].x = coord[0].x - 2;
		temp_coord[3].y = coord[0].y - 1;
	} else if (dir == west) {
		temp_coord[0].x = coord[0].x - 2;
		temp_coord[0].y = coord[0].y;
		temp_coord[1].x = coord[0].x - 2;
		temp_coord[1].y = coord[0].y + 1;
		temp_coord[2].x = coord[0].x - 1;
		temp_coord[2].y = coord[0].y + 1;
		temp_coord[3].x = coord[0].x - 1;
		temp_coord[3].y = coord[0].y + 2;
	} else {
		temp_coord[0].x = coord[0].x;
		temp_coord[0].y = coord[0].y + 2;
		temp_coord[1].x = coord[0].x + 1;
		temp_coord[1].y = coord[0].y + 2;
		temp_coord[2].x = coord[0].x + 1;
		temp_coord[2].y = coord[0].y + 1;
		temp_coord[3].x = coord[0].x + 2;
		temp_coord[3].y = coord[0].y + 1;
	}
}

static void rotate_t() {
	
	if (dir == east) {
		temp_coord[0].x = coord[0].x + 1;
		temp_coord[0].y = coord[0].y + 1;
		temp_coord[1].x = coord[0].x + 1;
		temp_coord[1].y = coord[0].y;
		temp_coord[2].x = coord[0].x + 1;
		temp_coord[2].y = coord[0].y - 1;
		temp_coord[3].x = coord[0].x + 2;
		temp_coord[3].y = coord[0].y;
	} else if (dir == north) {
		temp_coord[0].x = coord[0].x + 1;
		temp_coord[0].y = coord[0].y - 1;
		temp_coord[1].x = coord[0].x;
		temp_coord[1].y = coord[0].y - 1;
		temp_coord[2].x = coord[0].x - 1;
		temp_coord[2].y = coord[0].y - 1;
		temp_coord[3].x = coord[0].x;
		temp_coord[3].y = coord[0].y - 2;
	} else if (dir == west) {
		temp_coord[0].x = coord[0].x - 1;
		temp_coord[0].y = coord[0].y - 1;
		temp_coord[1].x = coord[0].x - 1;
		temp_coord[1].y = coord[0].y;
		temp_coord[2].x = coord[0].x - 1;
		temp_coord[2].y = coord[0].y + 1;
		temp_coord[3].x = coord[0].x - 2;
		temp_coord[3].y = coord[0].y;
	} else {
		temp_coord[0].x = coord[0].x - 1;
		temp_coord[0].y = coord[0].y + 1;
		temp_coord[1].x = coord[0].x;
		temp_coord[1].y = coord[0].y + 1;
		temp_coord[2].x = coord[0].x + 1;
		temp_coord[2].y = coord[0].y + 1;
		temp_coord[3].x = coord[0].x;
		temp_coord[3].y = coord[0].y + 2;
	}
}

static void rotate_z() {

	if (dir == east) {
		temp_coord[0].x = coord[0].x + 1;
		temp_coord[0].y = coord[0].y + 1;
		temp_coord[1].x = coord[0].x + 1;
		temp_coord[1].y = coord[0].y;
		temp_coord[2].x = coord[0].x + 2;
		temp_coord[2].y = coord[0].y;
		temp_coord[3].x = coord[0].x + 2;
		temp_coord[3].y = coord[0].y - 1;
	} else if (dir == north) {
		temp_coord[0].x = coord[0].x + 1;
		temp_coord[0].y = coord[0].y - 1;
		temp_coord[1].x = coord[0].x;
		temp_coord[1].y = coord[0].y - 1;
		temp_coord[2].x = coord[0].x;
		temp_coord[2].y = coord[0].y - 2;
		temp_coord[3].x = coord[0].x - 1;
		temp_coord[3].y = coord[0].y - 2;
	} else if (dir == west) {
		temp_coord[0].x = coord[0].x - 1;
		temp_coord[0].y = coord[0].y - 1;
		temp_coord[1].x = coord[0].x - 1;
		temp_coord[1].y = coord[0].y;
		temp_coord[2].x = coord[0].x - 2;
		temp_coord[2].y = coord[0].y;
		temp_coord[3].x = coord[0].x - 2;
		temp_coord[3].y = coord[0].y + 1;
	} else {
		temp_coord[0].x = coord[0].x - 1;
		temp_coord[0].y = coord[0].y + 1;
		temp_coord[1].x = coord[0].x;
		temp_coord[1].y = coord[0].y + 1;
		temp_coord[2].x = coord[0].x;
		temp_coord[2].y = coord[0].y + 2;
		temp_coord[3].x = coord[0].x + 1;
		temp_coord[3].y = coord[0].y + 2;
	}
}

static void rotate_piece() {
	
	if (piece == i)
		rotate_i();
	else if (piece == j)
		rotate_j();
	else if (piece == l)
		rotate_l();
	else if (piece == s)
		rotate_s();
	else if (piece == t)
		rotate_t();
	else if (piece == z)
		rotate_z();
	if (collision())
		return;
	update_coord();
	update_dir();
}

static void collapse_pieces(int line) {
	
	int j;
	for (j = 0; j < X; ++j)
		map[line][j] = false;
	for (int i = line; i > 0; --i) {
		for (j = 0; j < X; ++j)
			map[i][j] = map[i - 1][j];
	}
	for (j = 0; j < X; ++j)
		map[0][j] = false;
}

static void try_to_collapse_pieces() {
	
	for (int i = 0; i < Y; ++i) {
		bool filled = true;
		for (int j = 0; j < X && filled; ++j)
			if (!map[i][j])
				filled = false;
		if (filled) {
			collapse_pieces(i);
			++score;
		}
	}
}

static void try_game_over() {
	
	for (int i = 0; i < 4 && !game_over; ++i)
		if ((piece == o && !coord[i].y) || (coord[i].y == 1))
			game_over = 1;
}

static void update_piece_coord() {
	
	if (piece == no_piece) {
		try_to_collapse_pieces();
		init_piece();
		return;
	}
	for (int i = 0; i < 4; ++i) {
		temp_coord[i].x = coord[i].x;
		temp_coord[i].y = coord[i].y + 1;
	}
	if (collision()) {
		try_game_over();
		piece = no_piece;
		return;
	}
	update_coord();
	if (button == right)
		move_piece_to_right();
	if (button == left)
		move_piece_to_left();
	if (button == rotate)
		rotate_piece();
	if (button == speed_up)
		fall += 1;
	button = unpressed;
}

int main() {
	
	init();
	nokia_lcd_init();
	srand(time(NULL));
	while (1) {
		nokia_lcd_clear();
		get_pressed_button();
		if (!game_started) {
			if (button == unpressed) {
				start_lcd();
			} else if (button == start && level_choosed == 0) {
				chose_level();
				level_choosed = 1;
			} else if(level_choosed == 1 && button == start) {
				game_started = 1;
			} else {
				button = unpressed;
			}
		} else {
			update_piece_coord();
			if (!game_over)
					display_map();
			else
				game_over_lcd();
		}
		nokia_lcd_render();
		if (game_over) {
			reset();
		} else {
			if (game_started) {
				if (fall) {
					_delay_ms(100);
					fall = 0;
				} else {
					if (level == 1)
						_delay_ms(LEVEL_1_DELAY);
					else
						_delay_ms(LEVEL_2_DELAY);
				}
			} else {
				_delay_ms(2000);
			}
		}
	}
	return 0;
}